%{
% Originally by Jakob Verbeek, Intelligent Systems Laboratory Amsterdam, 
% University of Amsterdam, 2005.
% http://www.science.uva.nl/~jverbeek

Modified by Charlie Tang with the more stable "NEWFORMULATION"
===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================
%}

function [ LogL, proj, lat_covs ] = ll_mfa(X, mfa, X2, NEWFORMULATION)
%
% function [ LogL, proj ] = mppca_gauss(X,M,W,Psi);
%
% computes log-likelihoods of X under Gaussians
% also returns the inferred latent coordinates
%
% X (D x N)     - n d-dimensional data
% mfa is a struct with the fields:
% M (D x C)     - k d-dimensional means
% W (D x d x C) - for each Gaussian q d-dim vectors
% Psi (D,C)     - diagonal noise levels
%
% LogL is C by N matrix
% proj is d by N by C matrix of projections
% lat_covs is d by d by C


N         = size(X,2);
[D, d, C] = size(mfa.W);

LogL = zeros(C,N);

if nargout > 1
    proj = zeros(d,N,C);
end


if nargout > 2
    lat_covs = zeros(d,d,C);
end

if ~exist('X2', 'var') || isempty(X2)
    X2 = X.^2;
end

if ~exist('NEWFORMULATION', 'var')
    NEWFORMULATION = 1;
end

if NEWFORMULATION == 0
   warning('ll_mfa, NEWFORMULATION == 0');
end

for c = 1:C
    w             = mfa.W(:,:,c);
    psi           = mfa.Psi(:,c);
    
    psiI          = psi.^-1;
    psiIw         = bsxfun(@times, w, psiI);
    
    if NEWFORMULATION==1 %experimental code, do not invert explicitly        
        lat_precision = eye(d) + w'* psiIw; % V_c
        %psiI_w_xc     = psiIw' * X - repmat(psiIw'*mfa.M(:,c),1,N);
        diff = bsxfun(@minus, X, mfa.M(:,c));
        psiI_w_xc     = psiIw'*diff;
        if nargout > 1
            proj(:,:,c)   = lat_precision \ psiI_w_xc;
        end

        det_latcov = 1./det(lat_precision);
        if det_latcov == 0
            L=chol(lat_precision);
            log_det_lat_cov = -2*sum(log(diag(L)));
        else
            log_det_lat_cov = log( det_latcov );
            %assert( abs(log(det(lat_cov)) - log_det_lat_cov) < 1e-5);
        end

        log_det = sum(log(psi)) - log_det_lat_cov;

        %energy = -.5*( psiI'*X2 + psiI'*(mfa.M(:,c).^2)-2*(psiI.*mfa.M(:,c))'*X) ...
        %        +.5 * sum( psiI_w_xc .* ( lat_precision \ psiI_w_xc),1);
        
        energy = -.5*psiI'*(diff.*diff) +.5 * sum( psiI_w_xc .* ( lat_precision \ psiI_w_xc),1);
        LogL(c,:) = energy -.5*log_det -(D/2)*log(2*pi);
        
        if nargout > 2
            lat_cov   = inv( lat_precision ); %cov of p(z|x), or V_c^-1        
        end    
        
    else %standard code
        
        lat_cov       = inv( eye(d) + w'* psiIw ); %cov of p(z|x), or V_c^-1
        psiI_w_xc     = psiIw' * X - repmat(psiIw'*mfa.M(:,c),1,N);

        if nargout > 1
           proj(:,:,c)   = lat_cov * psiI_w_xc; 
        end

        det_latcov = det(lat_cov);
        if det_latcov == 0
            L=chol(lat_cov);
            log_det_lat_cov = 2*sum(log(diag(L)));
        else
            log_det_lat_cov = log( det_latcov );
            %assert( abs(log(det(lat_cov)) - log_det_lat_cov) < 1e-5);
        end

        log_det = sum(log(psi)) - log_det_lat_cov;

        energy = -.5*( psiI'*X2 + psiI'*(mfa.M(:,c).^2)-2*(psiI.*mfa.M(:,c))'*X) ...
                 +.5 * sum( psiI_w_xc .* (lat_cov*psiI_w_xc),1);

        LogL(c,:) = energy -.5*log_det -(D/2)*log(2*pi);
    end
    
    if nargout > 2    
        lat_covs(:,:,c) = lat_cov; 
    end    
end

assert(all(isfinite(LogL(:))));
