%{

===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
CT 3/2012

Modified to work with more than 2 sets of latent factors
M-step of EM for Mixture of Tensor Analyzers

PURPOSE:
INPUT:     params.iso - for each mixture component have equal Psi(i)
           params.eq  - all components have same Psi
           params.GPU - 0 use CPU, 1 use GPU for matrix multiplication
           
           params.QTHRESH - if the average (over samples) posterior
                    probability of a component is less than this thresh
                    we don't update the parameters of that component
           params.MIN_PSI - no Psi value can be less than this
        
        MoTA is a struct with the fields:
             M (D x C)     - k d-dimensional means
             W (D x (d1+d2) x C) - 2-biases
             Psi (D,C)     - diagonal noise levels
             T - tensor (matricized) D x (d1*d2) x C
             mix - C x 1 mixing proportions             
             .d - row vector of latent dimensions
        
        X - D x N data matrix
        Q - C x N  posterior probabilities
        zs{d,c} is a 2d cell of samples, each cell has dim by N samples.        
        
OUTPUT:
NOTES:
TESTED:
CHANGELOG: modified from original verbeek code
TODO:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%}

function [mta] = mta3_M_step( params, X, Q, zs, MoTA, nUpdateIteratons, ...
    ZM, learnrate, wtcost)

D       = size(X,1);
C = size(zs,1);
[d1,N] = size(zs{1,1});

assert( N == size(X,2) );

if ~exist('ZM', 'var') || isempty(ZM)
    ZM = 0;
end

% first copy over the previous parameters
mta = MoTA;

for c=1:C; % M-step
    
    sumQ = sum(Q(c,:)); % 1 by N posterior probabilities
    
    if sumQ/N > params.QTHRESH   %update the parameters

        y = [cat(1, zs{c,:}); ones(1,N)]; %d1+d2+1 x N
        %u = khatrirao(z2(:,:,c), z1(:,:,c));   %d1*d2 x N
        u = khatrirao( zs{c,:}, 'r');

        sumQy = bsxfun(@times, Q(c,:), y);
        Exyt = X*sumQy'; %D x d1+d2+1
        Euyt = u*sumQy'; %d1*d2 x d1+d2+1
        Eyyt = y*sumQy'; %d1+d2+1 x d1+d2+1
        
        sumQX = bsxfun(@times, Q(c,:), X);
        Exxt = X*sumQX';
        Euxt = u*sumQX';
        
        sumQu = bsxfun(@times, Q(c,:), u);
        Euut = u*sumQu';
        
        %may do a few iterations of loop instead
        for kk = 1:nUpdateIteratons,
            %W
            %tmp = mrdivide( Exyt-mta.T(:,:,c)*Euyt,  Eyyt); %D x d1+d2+1
            tmp =  (Exyt-mta.T(:,:,c)*Euyt)*pinv(Eyyt);
            
            %T
            %mta.T(:,:,c) = mrdivide( Euxt'-tmp*Euyt',  Euut); %D x d1*d2        
            T_new = ( Euxt'-tmp*Euyt')*pinv(Euut);
            mta.T(:,:,c) =  mta.T(:,:,c) + learnrate*(T_new- mta.T(:,:,c)) -learnrate*wtcost*mta.T(:,:,c);
        end
        
        %mta.W(:,:,c)    = tmp(:,1:end-1);
        %mta.M(:,c)      = tmp(:,end);
        W_new = tmp(:,1:end-1);
        M_new = tmp(:,end);
        
        mta.W(:,:,c) =  mta.W(:,:,c) + learnrate*(W_new- mta.W(:,:,c)) -learnrate*wtcost*mta.W(:,:,c);
        mta.M(:,c) = mta.M(:,c) + learnrate*(M_new-mta.M(:,c));
        
                
        %psiinv 
        Psi_new = diag( Exxt-2*mta.T(:,:,c)*(Euxt-Euyt*tmp'-0.5*Euut*mta.T(:,:,c)')-2*tmp*(Exyt'-0.5*Eyyt*tmp' ) )/sumQ;
        mta.Psi(:,c) = mta.Psi(:,c) + learnrate*(Psi_new-mta.Psi(:,c));
        
        mta.Psi(:,c)  = max(mta.Psi(:,c), params.MIN_PSI );
        if params.iso==1
            mta.Psi(:,c) = mean(mta.Psi(:,c));
        end
        
    end %responsibilities over a threshold
    
    %mix
    mta.mix(c,1)     = sumQ / N;
end


%other regularizations
if params.eq==1
    mta.Psi = (mta.Psi*mta.mix)*ones(1,C);
end

% hack give some components a little bit of mixture
smallmix = 0.1/C;
mta.mix = mta.mix+smallmix;
mta.mix = mta.mix/sum(mta.mix);

if ZM == 1
    mta.M(:) = 0;
end

